import os
import random
import traceback
import pandas as pd
from river import base
from tqdm.notebook import tqdm
from datetime import datetime, timezone
from IPython.display import display

from virny.configs.constants import ModelSetting
from virny.utils.protected_groups_partitioning import create_test_protected_groups
from virny.custom_classes.base_dataset import BaseFlowDataset
from virny.analyzers.subgroup_variance_analyzer import SubgroupVarianceAnalyzer
from virny.utils.common_helpers import save_metrics_to_file
from virny.analyzers.subgroup_error_analyzer import SubgroupErrorAnalyzer


def compute_model_metrics_with_config(base_model, model_name: str, dataset: BaseFlowDataset, config, save_results_dir_path: str,
                                      save_results: bool = True, verbose: int = 0) -> pd.DataFrame:
    """
    Compute subgroup metrics for the base model. Arguments are defined as an input config object.
    Save results in `save_results_dir_path` folder.

    Return a dataframe of model metrics.

    Parameters
    ----------
    base_model
        Base model for metrics computation
    model_name
        Model name to name a result file with metrics
    dataset
        BaseFlowDataset object that contains all needed attributes like target, features, numerical_columns etc.
    config
        Object that contains bootstrap_fraction, dataset_name, n_estimators, sensitive_attributes_dct attributes
    save_results_dir_path
        Location where to save result files with metrics
    save_results
        [Optional] If to save result metrics in a file
    verbose
        [Optional] Level of logs printing. The greater level provides more logs.
            As for now, 0, 1, 2 levels are supported.

    """
    return compute_model_metrics(base_model=base_model,
                                 n_estimators=config.n_estimators,
                                 dataset=dataset,
                                 bootstrap_fraction=config.bootstrap_fraction,
                                 sensitive_attributes_dct=config.sensitive_attributes_dct,
                                 dataset_name=config.dataset_name,
                                 base_model_name=model_name,
                                 save_results=save_results,
                                 save_results_dir_path=save_results_dir_path,
                                 verbose=verbose)


def compute_model_metrics(base_model, n_estimators: int, dataset: BaseFlowDataset, bootstrap_fraction: float,
                          sensitive_attributes_dct: dict, dataset_name: str, base_model_name: str,
                          model_setting: str = ModelSetting.BATCH.value, computation_mode: str = None, save_results: bool = True,
                          save_results_dir_path: str = None, verbose: int = 0):
    """
    Compute subgroup metrics for the base model.
    Save results in `save_results_dir_path` folder.

    Return a dataframe of model metrics.

    Parameters
    ----------
    base_model
        Base model for metrics computation
    n_estimators
        Number of estimators for bootstrap to compute subgroup variance metrics
    dataset
        BaseFlowDataset object that contains all needed attributes like target, features, numerical_columns etc.
    bootstrap_fraction
        Fraction of a train set in range [0.0 - 1.0] to fit models in bootstrap
    sensitive_attributes_dct
        A dictionary where keys are sensitive attribute names (including attributes intersections),
         and values are privilege values for these attributes
    dataset_name
        Dataset name to name a result file with metrics
    base_model_name
        Model name to name a result file with metrics
    save_results
        [Optional] If to save result metrics in a file
    model_setting
        [Optional] Model type: 'batch' or 'incremental'. Default: 'batch'.
    computation_mode
        [Optional] A non-default mode for metrics computation. Should be included in the ComputationMode enum.
    save_results_dir_path
        [Optional] Location where to save result files with metrics
    verbose
        [Optional] Level of logs printing. The greater level provides more logs.
            As for now, 0, 1, 2 levels are supported.

    """
    model_setting = ModelSetting.BATCH if model_setting is None else ModelSetting[model_setting.upper()]

    test_protected_groups = create_test_protected_groups(dataset.X_test, dataset.init_features_df, sensitive_attributes_dct)
    if verbose >= 2:
        print('\nProtected groups splits:')
        for g in test_protected_groups.keys():
            print(g, test_protected_groups[g].shape)

    # Compute stability metrics for subgroups
    subgroup_variance_analyzer = SubgroupVarianceAnalyzer(model_setting=model_setting,
                                                          n_estimators=n_estimators,
                                                          base_model=base_model,
                                                          base_model_name=base_model_name,
                                                          bootstrap_fraction=bootstrap_fraction,
                                                          dataset=dataset,
                                                          dataset_name=dataset_name,
                                                          sensitive_attributes_dct=sensitive_attributes_dct,
                                                          test_protected_groups=test_protected_groups,
                                                          computation_mode=computation_mode,
                                                          verbose=verbose)
    y_preds, variance_metrics_df = subgroup_variance_analyzer.compute_metrics(save_results=False,
                                                                              result_filename=None,
                                                                              save_dir_path=None,
                                                                              make_plots=False)

    # Compute error metrics for subgroups
    error_analyzer = SubgroupErrorAnalyzer(X_test=dataset.X_test,
                                           y_test=dataset.y_test,
                                           sensitive_attributes_dct=sensitive_attributes_dct,
                                           test_protected_groups=test_protected_groups,
                                           computation_mode=computation_mode)
    dtc_res = error_analyzer.compute_subgroup_metrics(y_preds,
                                                      save_results=False,
                                                      result_filename=None,
                                                      save_dir_path=None)
    error_metrics_df = pd.DataFrame(dtc_res)

    metrics_df = pd.concat([variance_metrics_df, error_metrics_df])
    metrics_df = metrics_df.reset_index()
    metrics_df = metrics_df.rename(columns={"index": "Metric"})
    metrics_df['Model_Name'] = base_model_name
    if isinstance(base_model, base.Classifier): # skip for incremental models
        metrics_df['Model_Params'] = None
    else:
        metrics_df['Model_Params'] = str(base_model.get_params())

    if save_results:
        # Save metrics
        result_filename = f'Metrics_{dataset_name}_{base_model_name}'
        save_metrics_to_file(metrics_df, result_filename, save_results_dir_path)

    return metrics_df


def run_metrics_computation(dataset: BaseFlowDataset, bootstrap_fraction: float, dataset_name: str,
                            models_config: dict, n_estimators: int, sensitive_attributes_dct: dict,
                            model_setting: str = ModelSetting.BATCH.value, computation_mode: str = None,
                            save_results: bool = True, save_results_dir_path: str = None, verbose: int = 0) -> dict:
    """
    Compute stability and accuracy metrics for each model in models_config.
    Save results in `save_results_dir_path` folder.

    Return a dictionary where keys are model names, and values are metrics for sensitive attributes defined in config.

    Parameters
    ----------
    dataset
        Dataset object that contains all needed attributes like target, features, numerical_columns etc.
    bootstrap_fraction
        Fraction of a train set in range [0.0 - 1.0] to fit models in bootstrap
    dataset_name
        Dataset name to name a result file with metrics
    models_config
        Dictionary where keys are model names, and values are initialized models
    n_estimators
        Number of estimators for bootstrap to compute subgroup stability metrics
    sensitive_attributes_dct
        A dictionary where keys are sensitive attribute names (including attributes intersections),
         and values are privilege values for these attributes
    model_setting
        [Optional] Model type: 'batch' or incremental. Default: 'batch'.
    computation_mode
        [Optional] A non-default mode for metrics computation. Should be included in the ComputationMode enum.
    save_results
        [Optional] If to save result metrics in a file
    save_results_dir_path
        [Optional] Location where to save result files with metrics
    verbose
        [Optional] Level of logs printing. The greater level provides more logs.
            As for now, 0, 1, 2 levels are supported.

    """
    models_metrics_dct = dict()
    num_models = len(models_config)
    for model_idx, model_name in tqdm(enumerate(models_config.keys()),
                                      total=num_models,
                                      desc="Analyze models in one run",
                                      colour="red"):
        if verbose >= 1:
            print('#' * 30, f' [Model {model_idx + 1} / {num_models}] Analyze {model_name} ', '#' * 30)
        try:
            base_model = models_config[model_name]
            model_metrics_df = compute_model_metrics(base_model=base_model,
                                                     n_estimators=n_estimators,
                                                     dataset=dataset,
                                                     bootstrap_fraction=bootstrap_fraction,
                                                     sensitive_attributes_dct=sensitive_attributes_dct,
                                                     model_setting=model_setting,
                                                     computation_mode=computation_mode,
                                                     dataset_name=dataset_name,
                                                     base_model_name=model_name,
                                                     save_results=save_results,
                                                     save_results_dir_path=save_results_dir_path,
                                                     verbose=verbose)
            models_metrics_dct[model_name] = model_metrics_df
            if verbose >= 2:
                print(f'\n[{model_name}] Metrics matrix:')
                display(model_metrics_df)
        except Exception as err:
            print('#' * 20, f'ERROR with {model_name}', '#' * 20)
            traceback.print_exc()

        if verbose >= 1:
            print('\n\n\n')

    return models_metrics_dct


def compute_metrics_with_config(dataset: BaseFlowDataset, config, models_config: dict,
                                save_results_dir_path: str, verbose: int = 0) -> dict:
    """
    Compute stability and accuracy metrics for each model in models_config. Arguments are defined as an input config object.
    Save results in `save_results_dir_path` folder.

    Return a dictionary where keys are model names, and values are metrics for sensitive attributes defined in config.

    Parameters
    ----------
    dataset
        BaseFlowDataset object that contains all needed attributes like target, features, numerical_columns etc.
    config
        Object that contains bootstrap_fraction, dataset_name, n_estimators, sensitive_attributes_dct attributes
    models_config
        Dictionary where keys are model names, and values are initialized models
    save_results_dir_path
        Location where to save result files with metrics
    verbose
        [Optional] Level of logs printing. The greater level provides more logs.
            As for now, 0, 1, 2 levels are supported.

    """
    start_datetime = datetime.now(timezone.utc)
    os.makedirs(save_results_dir_path, exist_ok=True)

    model_metrics_dct = dict()
    models_metrics_dct = run_metrics_computation(dataset=dataset,
                                                 bootstrap_fraction=config.bootstrap_fraction,
                                                 dataset_name=config.dataset_name,
                                                 models_config=models_config,
                                                 n_estimators=config.n_estimators,
                                                 sensitive_attributes_dct=config.sensitive_attributes_dct,
                                                 model_setting=config.model_setting,
                                                 computation_mode=config.computation_mode,
                                                 save_results=False,
                                                 verbose=verbose)

    # Concatenate with previous results and save them in an overwrite mode each time for backups
    for model_name in models_metrics_dct.keys():
        model_metrics_df = models_metrics_dct[model_name]
        model_metrics_dct[model_name] = model_metrics_df

        result_filename = f'Metrics_{config.dataset_name}_{model_name}_{config.n_estimators}_Estimators_{start_datetime.strftime("%Y%m%d__%H%M%S")}.csv'
        model_metrics_dct[model_name].to_csv(f'{save_results_dir_path}/{result_filename}', index=False, mode='w')

    return model_metrics_dct


def compute_metrics_multiple_runs_with_db_writer(dataset: BaseFlowDataset, config, models_config: dict,
                                                 custom_tbl_fields_dct: dict, db_writer_func, verbose: int = 0) -> dict:
    """
    Compute stability and accuracy metrics for each model in models_config. Arguments are defined as an input config object.
    Save results to a database after each run appending fields and value from custom_tbl_fields_dct and using db_writer_func.

    Return a dictionary where keys are model names, and values are metrics for sensitive attributes defined in config.

    Parameters
    ----------
    dataset
        BaseFlowDataset object that contains all needed attributes like target, features, numerical_columns etc.
    config
        Object that contains bootstrap_fraction, dataset_name, n_estimators, sensitive_attributes_dct attributes
    models_config
        Dictionary where keys are model names, and values are initialized models
    custom_tbl_fields_dct
        Dictionary where keys are column names and values to add to inserted metrics during saving results to a database
    db_writer_func
        Python function object has one argument (run_models_metrics_df) and save this metrics df to a target database
    verbose
        [Optional] Level of logs printing. The greater level provides more logs.
            As for now, 0, 1, 2 levels are supported.

    """
    multiple_runs_metrics_dct = dict()
    run_models_metrics_df = pd.DataFrame()
    models_metrics_dct = run_metrics_computation(dataset=dataset,
                                                 bootstrap_fraction=config.bootstrap_fraction,
                                                 dataset_name=config.dataset_name,
                                                 models_config=models_config,
                                                 n_estimators=config.n_estimators,
                                                 sensitive_attributes_dct=config.sensitive_attributes_dct,
                                                 model_setting=config.model_setting,
                                                 computation_mode=config.computation_mode,
                                                 save_results=False,
                                                 verbose=verbose)

    # Concatenate current run metrics with previous results and
    # create melted_model_metrics_df to save it in a database
    for model_name in models_metrics_dct.keys():
        model_metrics_df = models_metrics_dct[model_name]
        model_metrics_df['Dataset_Name'] = config.dataset_name
        model_metrics_df['Num_Estimators'] = config.n_estimators

        model_metrics_df_copy = model_metrics_df.copy(deep=True)  # Version copy for multiple_runs_metrics_dct
        # Append current run metrics to multiple_runs_metrics_dct
        if multiple_runs_metrics_dct.get(model_name) is None:
            multiple_runs_metrics_dct[model_name] = model_metrics_df_copy
        else:
            multiple_runs_metrics_dct[model_name] = pd.concat([multiple_runs_metrics_dct[model_name], model_metrics_df_copy])

        # Extend df with technical columns
        model_metrics_df['Tag'] = 'OK'
        model_metrics_df['Record_Create_Date_Time'] = datetime.now(timezone.utc)
        for column, value in custom_tbl_fields_dct.items():
            model_metrics_df[column] = value

        subgroup_names = [col for col in model_metrics_df.columns if '_priv' in col or '_dis' in col] + ['overall']
        melted_model_metrics_df = model_metrics_df.melt(id_vars=[col for col in model_metrics_df.columns if col not in subgroup_names],
                                                        value_vars=subgroup_names,
                                                        var_name="Subgroup",
                                                        value_name="Metric_Value")
        run_models_metrics_df = pd.concat([run_models_metrics_df, melted_model_metrics_df])

    # Save results for this run in a database
    db_writer_func(run_models_metrics_df)

    return multiple_runs_metrics_dct


def compute_metrics_multiple_runs_with_multiple_test_sets(dataset: BaseFlowDataset, extra_test_sets_lst,
                                                          config, models_config: dict, custom_tbl_fields_dct: dict,
                                                          db_writer_func, verbose: int = 0):
    """
    Compute stability and accuracy metrics for each model in models_config based on dataset.X_test and each extra test set
     in extra_test_sets_lst. Arguments are defined as an input config object. Save results to a database after each run
      appending fields and value from custom_tbl_fields_dct and using db_writer_func.
      Index of each test set is also added as a separate column in out final records in the database
      (0 index -- for dataset.X_test, 1 and greater -- for each extra test set in extra_test_sets_lst, keeping the original sequence).

    Parameters
    ----------
    dataset
        BaseFlowDataset object that contains all needed attributes like target, features, numerical_columns etc.
    extra_test_sets_lst
        List of extra test sets like [(X_test1, y_test1), (X_test2, y_test2), ...] to compute metrics
        that are not equal to original dataset.X_test and dataset.y_test
    config
        Object that contains bootstrap_fraction, dataset_name, n_estimators, sensitive_attributes_dct attributes
    models_config
        Dictionary where keys are model names, and values are initialized models
    custom_tbl_fields_dct
        Dictionary where keys are column names and values to add to inserted metrics during saving results to a database
    db_writer_func
        Python function object has one argument (run_models_metrics_df) and save this metrics df to a target database
    verbose
        [Optional] Level of logs printing. The greater level provides more logs.
            As for now, 0, 1, 2 levels are supported.

    """
    models_metrics_dct = run_metrics_computation_with_multiple_test_sets(dataset=dataset,
                                                                         bootstrap_fraction=config.bootstrap_fraction,
                                                                         dataset_name=config.dataset_name,
                                                                         extra_test_sets_lst=extra_test_sets_lst,
                                                                         models_config=models_config,
                                                                         n_estimators=config.n_estimators,
                                                                         sensitive_attributes_dct=config.sensitive_attributes_dct,
                                                                         model_setting=config.model_setting,
                                                                         computation_mode=config.computation_mode,
                                                                         verbose=verbose)

    # Concatenate current run metrics with previous results and
    # create melted_model_metrics_df to save it in a database
    run_models_metrics_df = pd.DataFrame()
    for model_name in models_metrics_dct.keys():
        model_metrics_dfs_lst = models_metrics_dct[model_name]
        for idx, model_metrics_df in enumerate(model_metrics_dfs_lst):
            model_metrics_df['Dataset_Name'] = config.dataset_name
            model_metrics_df['Num_Estimators'] = config.n_estimators
            model_metrics_df['Test_Set_Index'] = idx

            # Extend df with technical columns
            model_metrics_df['Tag'] = 'OK'
            model_metrics_df['Record_Create_Date_Time'] = datetime.now(timezone.utc)
            for column, value in custom_tbl_fields_dct.items():
                model_metrics_df[column] = value

            subgroup_names = [col for col in model_metrics_df.columns if '_priv' in col or '_dis' in col] + ['overall']
            melted_model_metrics_df = model_metrics_df.melt(id_vars=[col for col in model_metrics_df.columns if col not in subgroup_names],
                                                            value_vars=subgroup_names,
                                                            var_name="Subgroup",
                                                            value_name="Metric_Value")
            run_models_metrics_df = pd.concat([run_models_metrics_df, melted_model_metrics_df])

    # Save results for this run in a database
    db_writer_func(run_models_metrics_df)

    if verbose >= 1:
        print('Metrics computation interface was successfully executed!')


def run_metrics_computation_with_multiple_test_sets(dataset: BaseFlowDataset, bootstrap_fraction: float, dataset_name: str,
                                                    extra_test_sets_lst: list, models_config: dict, n_estimators: int,
                                                    sensitive_attributes_dct: dict, model_setting: str = ModelSetting.BATCH.value,
                                                    computation_mode: str = None, verbose: int = 0) -> dict:
    """
    Compute stability and accuracy metrics for each model in models_config based on dataset.X_test and each extra test set
     in extra_test_sets_lst. Save results in `save_results_dir_path` folder.

    Return a dictionary where keys are model names, and values are metrics for sensitive attributes defined in config.

    Parameters
    ----------
    dataset
        Dataset object that contains all needed attributes like target, features, numerical_columns etc.
    bootstrap_fraction
        Fraction of a train set in range [0.0 - 1.0] to fit models in bootstrap
    dataset_name
        Dataset name to name a result file with metrics
    extra_test_sets_lst
        List of extra test sets like [(X_test1, y_test1), (X_test2, y_test2), ...] to compute metrics
    models_config
        Dictionary where keys are model names, and values are initialized models
    n_estimators
        Number of estimators for bootstrap to compute subgroup stability metrics
    sensitive_attributes_dct
        A dictionary where keys are sensitive attribute names (including attributes intersections),
         and values are privilege values for these attributes
    model_setting
        Model type: 'batch' or incremental.
    computation_mode
        [Optional] A non-default mode for metrics computation. Should be included in the ComputationMode enum.
    verbose
        [Optional] Level of logs printing. The greater level provides more logs.
            As for now, 0, 1, 2 levels are supported.

    """
    models_metrics_dct = dict()
    num_models = len(models_config)
    for model_idx, model_name in tqdm(enumerate(models_config.keys()),
                                      total=num_models,
                                      desc="Analyze models in one run",
                                      colour="red"):
        if verbose >= 1:
            print('#' * 30, f' [Model {model_idx + 1} / {num_models}] Analyze {model_name} ', '#' * 30)
        try:
            base_model = models_config[model_name]
            model_metrics_dfs_lst = compute_model_metrics_with_multiple_test_sets(base_model=base_model,
                                                                                  n_estimators=n_estimators,
                                                                                  dataset=dataset,
                                                                                  extra_test_sets_lst=extra_test_sets_lst,
                                                                                  bootstrap_fraction=bootstrap_fraction,
                                                                                  sensitive_attributes_dct=sensitive_attributes_dct,
                                                                                  model_setting=model_setting,
                                                                                  computation_mode=computation_mode,
                                                                                  dataset_name=dataset_name,
                                                                                  base_model_name=model_name,
                                                                                  verbose=verbose)
            models_metrics_dct[model_name] = model_metrics_dfs_lst
        except Exception as err:
            print('#' * 20, f'ERROR with {model_name}', '#' * 20)
            traceback.print_exc()

        if verbose >= 1:
            print('\n\n\n')

    return models_metrics_dct


def compute_model_metrics_with_multiple_test_sets(base_model, n_estimators: int,
                                                  dataset: BaseFlowDataset, extra_test_sets_lst: list,
                                                  bootstrap_fraction: float, sensitive_attributes_dct: dict,
                                                  dataset_name: str, base_model_name: str,
                                                  model_setting: str = ModelSetting.BATCH.value,
                                                  computation_mode: str = None, verbose: int = 0):
    """
    Compute subgroup metrics for the base model based on dataset.X_test and each extra test set in extra_test_sets_lst.
    Save results in `save_results_dir_path` folder.

    Return a dataframe of model metrics.

    Parameters
    ----------
    base_model
        Base model for metrics computation
    n_estimators
        Number of estimators for bootstrap to compute subgroup stability metrics
    dataset
        BaseFlowDataset object that contains all needed attributes like target, features, numerical_columns etc.
    extra_test_sets_lst
        List of extra test sets like [(X_test1, y_test1), (X_test2, y_test2), ...] to compute metrics
    bootstrap_fraction
        Fraction of a train set in range [0.0 - 1.0] to fit models in bootstrap
    sensitive_attributes_dct
        A dictionary where keys are sensitive attribute names (including attributes intersections),
         and values are privilege values for these attributes
    dataset_name
        Dataset name to name a result file with metrics
    base_model_name
        Model name to name a result file with metrics
    model_setting
        Model type: 'batch' or incremental.
    computation_mode
        [Optional] A non-default mode for metrics computation. Should be included in the ComputationMode enum.
    verbose
        [Optional] Level of logs printing. The greater level provides more logs.
            As for now, 0, 1, 2 levels are supported.

    """
    model_setting = ModelSetting.BATCH if model_setting is None else ModelSetting[model_setting.upper()]
    subgroup_variance_analyzer = SubgroupVarianceAnalyzer(model_setting=model_setting,
                                                          n_estimators=n_estimators,
                                                          base_model=base_model,
                                                          base_model_name=base_model_name,
                                                          bootstrap_fraction=bootstrap_fraction,
                                                          dataset=dataset,  # will be replaced in the below for-cycle
                                                          dataset_name=dataset_name,
                                                          sensitive_attributes_dct=sensitive_attributes_dct,
                                                          test_protected_groups=dict(),  # stub for this attribute
                                                          computation_mode=computation_mode,
                                                          verbose=verbose)

    test_sets_lst = [(dataset.X_test, dataset.y_test)] + extra_test_sets_lst
    all_test_sets_metrics_lst = []
    for set_idx, (new_X_test, new_y_test) in enumerate(test_sets_lst):
        new_test_protected_groups = create_test_protected_groups(new_X_test, dataset.init_features_df, sensitive_attributes_dct)
        if verbose >= 2:
            print(f'\nProtected groups splits for test set index #{set_idx}:')
            for g in new_test_protected_groups.keys():
                print(g, new_test_protected_groups[g].shape)

        # Replace test sets and protected groups for each new test set
        subgroup_variance_analyzer.set_test_sets(new_X_test, new_y_test)
        subgroup_variance_analyzer.set_test_protected_groups(new_test_protected_groups)

        # Compute stability metrics for subgroups
        y_preds, variance_metrics_df = subgroup_variance_analyzer.compute_metrics(save_results=False,
                                                                                  result_filename=None,
                                                                                  save_dir_path=None,
                                                                                  make_plots=False,
                                                                                  with_fit=True if set_idx == 0 else False)

        # Compute accuracy metrics for subgroups
        error_analyzer = SubgroupErrorAnalyzer(X_test=new_X_test,
                                               y_test=new_y_test,
                                               sensitive_attributes_dct=sensitive_attributes_dct,
                                               test_protected_groups=new_test_protected_groups,
                                               computation_mode=computation_mode)
        dtc_res = error_analyzer.compute_subgroup_metrics(y_preds,
                                                          save_results=False,
                                                          result_filename=None,
                                                          save_dir_path=None)
        error_metrics_df = pd.DataFrame(dtc_res)

        metrics_df = pd.concat([variance_metrics_df, error_metrics_df])
        metrics_df = metrics_df.reset_index()
        metrics_df = metrics_df.rename(columns={"index": "Metric"})
        metrics_df['Model_Name'] = base_model_name
        metrics_df['Model_Params'] = str(base_model.get_params())

        all_test_sets_metrics_lst.append(metrics_df)

    return all_test_sets_metrics_lst



"""
Common helpers and utils.
"""
from .common_helpers import validate_config
from .stability_utils import count_prediction_stats
from .protected_groups_partitioning import create_test_protected_groups


__all__ = [
    "validate_config",
    "create_test_protected_groups",
    "count_prediction_stats",
]



"""
User interfaces.

This module contains user interfaces for metrics computation.
"""

from .metrics_computation_interfaces import (
    compute_model_metrics,
    compute_model_metrics_with_config,
    run_metrics_computation,
    compute_metrics_with_config,
    compute_metrics_multiple_runs_with_multiple_test_sets,
    compute_metrics_multiple_runs_with_db_writer
)


__all__ = [
    "compute_metrics_with_config",
    "compute_metrics_multiple_runs_with_multiple_test_sets",
    "compute_metrics_multiple_runs_with_db_writer",
    "compute_model_metrics",
    "compute_model_metrics_with_config",
    "run_metrics_computation",
]



import numpy as np
import pandas as pd

from virny.configs.constants import ComputationMode
from virny.utils.stability_utils import count_prediction_stats, combine_bootstrap_predictions
from virny.analyzers.abstract_subgroup_analyzer import AbstractSubgroupAnalyzer


class SubgroupVarianceCalculator(AbstractSubgroupAnalyzer):
    """
    Calculator that calculates variance metrics for subgroups.

    Parameters
    ----------
    X_test
        Processed features test set
    y_test
        Targets test set
    sensitive_attributes_dct
        A dictionary where keys are sensitive attributes names (including attributes intersections),
         and values are privilege values for these subgroups
    test_protected_groups
        A dictionary where keys are sensitive attributes, and values input dataset rows
         that are correspondent to these sensitive attributes.
    computation_mode
        [Optional] A non-default mode for metrics computation. Should be included in the ComputationMode enum.

    """
    def __init__(self, X_test: pd.DataFrame, y_test: pd.DataFrame, sensitive_attributes_dct: dict,
                 test_protected_groups=None, computation_mode: str = None):
        super().__init__(X_test, y_test, sensitive_attributes_dct, test_protected_groups, computation_mode)
        self.overall_variance_metrics = None
        self.subgroup_variance_metrics_dict = None

    def set_overall_variance_metrics(self, overall_variance_metrics):
        self.overall_variance_metrics = overall_variance_metrics

    def _partition_and_compute_metrics(self, models_predictions, results: dict):
        for group_name in self.test_protected_groups.keys():
            X_test_group = self.test_protected_groups[group_name]
            group_models_predictions = {
                model_idx: models_predictions[model_idx][X_test_group.index].reset_index(drop=True)
                for model_idx in models_predictions.keys()
            }
            results[group_name] = self._compute_metrics(self.y_test[X_test_group.index].reset_index(drop=True),
                                                        group_models_predictions)

        return results

    def _partition_and_compute_metrics_for_error_analysis(self, models_predictions, results: dict):
        """
        Partition predictions on correct and incorrect and compute subgroup metrics for each of the partitions.
        Used for the 'error_analysis' mode.

        :param models_predictions: a list of predictions
        :param results: a dict to add subgroup metrics for each partition
        """
        # Create a 1D pandas series of predictions for the test set based on bootstrap predictions
        y_preds = combine_bootstrap_predictions(models_predictions, self.y_test.index)

        # Partition and compute subgroup metrics
        for group_name in self.test_protected_groups.keys():
            X_test_group = self.test_protected_groups[group_name]
            group_y_true = self.y_test[X_test_group.index]
            group_y_preds = y_preds[X_test_group.index]

            # Define indexes of each partition of the group: overall group indexes,
            # correct preds group indexes, incorrect preds group indexes
            correct_preds_indexes = group_y_true.index[group_y_true == group_y_preds]
            incorrect_preds_indexes = group_y_true.index[group_y_true != group_y_preds]
            partition_indexes_dct = {
                group_name: X_test_group.index,
                f'{group_name}_correct': correct_preds_indexes,
                f'{group_name}_incorrect': incorrect_preds_indexes,
            }

            # Compute metrics for each group partition
            for group_partition_name, partition_indexes in partition_indexes_dct.items():
                group_models_predictions = {
                    model_idx: models_predictions[model_idx][partition_indexes].reset_index(drop=True)
                    for model_idx in models_predictions.keys()
                }
                metrics_dct = self._compute_metrics(self.y_test[partition_indexes].reset_index(drop=True),
                                                    group_models_predictions)
                results[group_partition_name] = metrics_dct

        return results

    def _compute_metrics(self, y_test: pd.DataFrame, group_models_predictions):
        _, _, prediction_stats = count_prediction_stats(y_test, group_models_predictions)
        return {
            'Jitter': prediction_stats.jitter,
            'Mean': np.mean(prediction_stats.means_lst),
            'Std': np.mean(prediction_stats.stds_lst),
            'IQR': np.mean(prediction_stats.iqr_lst),
            'Aleatoric_Uncertainty': np.mean(prediction_stats.mean_ensemble_entropy_lst),
            'Overall_Uncertainty': np.mean(prediction_stats.overall_entropy_lst),
            'Statistical_Bias': np.mean(prediction_stats.statistical_bias_lst),
            'Per_Sample_Accuracy': np.mean(prediction_stats.per_sample_accuracy_lst),
            'Label_Stability': np.mean(prediction_stats.label_stability_lst),
        }

    def compute_subgroup_metrics(self, models_predictions: dict, save_results: bool,
                                 result_filename: str = None, save_dir_path: str = None):
        """
        Compute variance metrics for subgroups.

        Return a dict of dicts where key is 'overall' or a subgroup name, and value is a dict of metrics for this subgroup.

        Parameters
        ----------
        models_predictions
            Dict of lists where key is a model index, and value is a list of model predictions based on X_test set
        save_results
            If we need to save result metrics in a file
        result_filename
            [Optional] Filename for results to save
        save_dir_path
            [Optional] Location where to save the results file

        """
        models_predictions = {
            model_idx: pd.Series(models_predictions[model_idx], index=self.y_test.index)
            for model_idx in models_predictions.keys()
        }

        # Compute overall stability metrics
        results = dict()
        results['overall'] = self.overall_variance_metrics

        # Compute stability metrics for subgroups
        if self.computation_mode == ComputationMode.ERROR_ANALYSIS.value:
            results = self._partition_and_compute_metrics_for_error_analysis(models_predictions, results)
        else:
            results = self._partition_and_compute_metrics(models_predictions, results)

        self.subgroup_variance_metrics_dict = results
        if save_results:
            self.save_metrics_to_file(result_filename, save_dir_path)

        return self.subgroup_variance_metrics_dict



"""
Subgroup Error and Variance Analyzers.

This module contains fairness and stability analysing methods for defined subgroups.
The purpose of an analyzer is to analyse defined metrics for defined subgroups.
"""
from .abstract_overall_variance_analyzer import AbstractOverallVarianceAnalyzer
from .abstract_subgroup_analyzer import AbstractSubgroupAnalyzer
from .batch_overall_variance_analyzer import BatchOverallVarianceAnalyzer
from .subgroup_error_analyzer import SubgroupErrorAnalyzer
from .subgroup_variance_analyzer import SubgroupVarianceAnalyzer
from .subgroup_variance_calculator import SubgroupVarianceCalculator

__all__ = [
    "AbstractOverallVarianceAnalyzer",
    "AbstractSubgroupAnalyzer",
    "BatchOverallVarianceAnalyzer",
    "SubgroupErrorAnalyzer",
    "SubgroupVarianceAnalyzer",
    "SubgroupVarianceCalculator",
]



import numpy as np
import pandas as pd

from virny.custom_classes.incremental_pandas_dataset import IncrementalPandasDataset
from virny.analyzers.abstract_overall_variance_analyzer import AbstractOverallVarianceAnalyzer


class IncrementalOverallVarianceAnalyzer(AbstractOverallVarianceAnalyzer):
    """
    Analyzer to compute subgroup variance metrics for incremental learning models.

    Parameters
    ----------
    base_model
        Base model for stability measuring
    base_model_name
        Model name like 'HoeffdingTreeClassifier' or 'LogisticRegression'
    bootstrap_fraction
        [0-1], fraction from train_pd_dataset for fitting an ensemble of base models
    X_train
        Processed features train set
    y_train
        Targets train set
    X_test
        Processed features test set
    y_test
        Targets test set
    target_column
        Name of the target column
    dataset_name
        Name of dataset, used for correct results naming
    n_estimators
        Number of estimators in ensemble to measure base_model stability
    verbose
        [Optional] Level of logs printing. The greater level provides more logs.
         As for now, 0, 1, 2 levels are supported.

    """
    def __init__(self, base_model, base_model_name: str, bootstrap_fraction: float,
                 X_train: pd.DataFrame, y_train: pd.DataFrame, X_test: pd.DataFrame, y_test: pd.DataFrame,
                 target_column: str, dataset_name: str, n_estimators: int, verbose: int = 0):
        super().__init__(base_model=base_model,
                         base_model_name=base_model_name,
                         bootstrap_fraction=bootstrap_fraction,
                         X_train=X_train,
                         y_train=y_train,
                         X_test=X_test,
                         y_test=y_test,
                         dataset_name=dataset_name,
                         n_estimators=n_estimators,
                         verbose=verbose)
        self.target_column = target_column
        self.dataset_reader = IncrementalPandasDataset

        # Create converters for the train set to apply them for train incremental datasets
        train_df_for_types = X_train.astype('object')
        train_converters = {str(col): type(train_df_for_types.loc[train_df_for_types.index[0], col])
                                 for col in train_df_for_types}
        train_converters[self.target_column] = type(y_train.astype('object')[y_train.index[0]])
        self.train_converters = train_converters

    def _fit_model(self, classifier, X_train: np.ndarray, y_train: np.ndarray):
        """
        Fit an incremental classifier that is an instance of self.base_model
        """
        train_df = pd.DataFrame(X_train, columns=[key for key in self.train_converters.keys()
                                                  if key != self.target_column])
        train_df[self.target_column] = y_train
        train_dataset = self.dataset_reader(pd_dataset=train_df, target=self.target_column, converters=self.train_converters)
        for x, y_true in train_dataset:
            classifier.learn_one(x=x, y=y_true)

        return classifier

    def _batch_predict(self, classifier, X_test: pd.DataFrame):
        """
        Predict with the incremental classifier for X_test set.
        Return predictions.
        """
        predictions = []
        test_df_for_types = X_test.astype('object')
        converters = {col: type(test_df_for_types.loc[test_df_for_types.index[0], col]) for col in test_df_for_types}
        test_dataset = self.dataset_reader(pd_dataset=X_test, target=None, converters=converters)
        for x, _ in test_dataset:
            y_pred = classifier.predict_one(x)
            predictions.append(y_pred)

        return predictions

    def _batch_predict_proba(self, classifier, X_test: pd.DataFrame):
        """
        Predict with the incremental classifier for X_test set.
        Return predicted probabilities for each class for each test point.
        """
        # Create converters for the test set to apply them for an incremental dataset
        test_df_for_types = X_test.astype('object')
        converters = {col: type(test_df_for_types.loc[test_df_for_types.index[0], col]) for col in test_df_for_types}

        predictions = []
        test_dataset = self.dataset_reader(pd_dataset=X_test, target=None, converters=converters)
        for x, _ in test_dataset:
            predict_proba = classifier.predict_proba_one(x)
            y_pred = predict_proba[0]
            predictions.append(y_pred)

        return predictions



"""
This module contains sample datasets and data loaders.
The purpose is to provide sample datasets for functionality testing and show examples of data loaders (aka dataset classes).
"""
from .data_loaders import CompasWithoutSensitiveAttrsDataset, DiabetesDataset, CompasDataset, \
    ACSIncomeDataset, ACSEmploymentDataset, ACSMobilityDataset, ACSTravelTimeDataset, ACSPublicCoverageDataset, \
    RicciDataset, LawSchoolDataset


__all__ = [
    "CompasWithoutSensitiveAttrsDataset",
    "CompasDataset",
    "DiabetesDataset",
    "RicciDataset",
    "LawSchoolDataset",
    "ACSIncomeDataset",
    "ACSEmploymentDataset",
    "ACSMobilityDataset",
    "ACSTravelTimeDataset",
    "ACSPublicCoverageDataset",
]



import os
import pandas as pd
import seaborn as sns

from matplotlib import pyplot as plt


def set_size(w,h, ax=None):
    """ w, h: width, height in inches """
    if not ax: ax=plt.gca()
    l = ax.figure.subplotpars.left
    r = ax.figure.subplotpars.right
    t = ax.figure.subplotpars.top
    b = ax.figure.subplotpars.bottom
    figw = float(w)/(r-l)
    figh = float(h)/(t-b)
    ax.figure.set_size_inches(figw, figh)


def plot_generic(x, y, xlabel, ylabel, x_lim, y_lim, plot_title):
    sns.set_style("darkgrid")
    plt.figure(figsize=(20,10))
    plt.scatter(x, y)
    plt.xlim(0, x_lim)
    plt.ylim(0, y_lim)
    plt.xlabel(xlabel, fontsize=14)
    plt.ylabel(ylabel, fontsize=14)
    plt.title(plot_title, fontsize=20)
    plt.show()


def create_average_metrics_df(dataset_name, model_names, metrics_path):
    results_filenames = [filename for filename in os.listdir(metrics_path)]
    models_average_results_dct = dict()
    for model_name in model_names:
        model_results_filenames = [filename for filename in results_filenames if 'Average_Metrics' not in filename
                                   and dataset_name in filename
                                   and model_name in filename]

        if len(model_results_filenames) == 0:
            continue

        model_results_dfs = []
        for model_results_filename in model_results_filenames:
            model_results_df = pd.read_csv(f'{metrics_path}/{model_results_filename}')
            model_results_df.set_index('index', inplace = True)
            model_results_dfs.append(model_results_df)

        model_average_results_df = None
        for model_results_df in model_results_dfs:
            if model_average_results_df is None:
                model_average_results_df = model_results_df
            else:
                model_average_results_df += model_results_df

        model_average_results_df = model_average_results_df / len(model_results_dfs)
        models_average_results_dct[model_name] = model_average_results_df

        filename = f'Average_Metrics_{dataset_name}_{model_name}.csv'
        model_average_results_df.reset_index().to_csv(f'{metrics_path}/{filename}', index=False)
        print(f'File with average metrics for {model_name} is created')

    return models_average_results_dct


def visualize_fairness_metrics_for_prediction_metric(models_average_results_dct, x_metric, y_metrics: list):
    sns.set_style("darkgrid")
    x_lim = 0.5
    y_lim = 0.22
    priv_dis_pairs = [('SEX_RAC1P_priv', 'SEX_RAC1P_dis'),
                      ('SEX_priv', 'SEX_dis'),
                      ('RAC1P_priv', 'RAC1P_dis')]
    for y_metric in y_metrics:
        for fairness_metric_priv, fairness_metric_dis in priv_dis_pairs:
            display_fairness_plot(models_average_results_dct, x_metric, y_metric,
                                  fairness_metric_priv, fairness_metric_dis, x_lim, y_lim)


def display_fairness_plot(models_average_results_dct, x_metric, y_metric,
                          fairness_metric_priv, fairness_metric_dis, x_lim, y_lim):
    fig, ax = plt.subplots()
    set_size(15, 8, ax)

    # List of all markers -- https://matplotlib.org/stable/api/markers_api.html
    markers = ['o', '*', '|', '<', '>', '^', 'v', '1', 's', 'x', 'D', 'P', 'H']
    model_names = models_average_results_dct.keys()
    shapes = []
    for idx, model_name in enumerate(model_names):
        x_val = abs(models_average_results_dct[model_name][fairness_metric_priv].loc[x_metric] - \
                    models_average_results_dct[model_name][fairness_metric_dis].loc[x_metric])
        y_val = abs(models_average_results_dct[model_name][fairness_metric_priv].loc[y_metric] - \
                    models_average_results_dct[model_name][fairness_metric_dis].loc[y_metric])
        a = ax.scatter(x_val, y_val, marker=markers[idx], s=100)
        shapes.append(a)

    plt.axhline(y=0.0, color='r', linestyle='-')
    plt.xlabel(f'{x_metric} Difference')
    plt.ylabel(f'{y_metric} Difference')
    plt.xlim(-0.01, x_lim)
    plt.ylim(-0.01, y_lim)
    plt.title(f'{fairness_metric_priv}-{fairness_metric_dis} difference for {x_metric} and {y_metric}', fontsize=20)
    ax.legend(shapes, model_names, fontsize=12, title='Markers')

    plt.show()



import os
import pathlib
import pkg_resources

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open
from os import path


NAME = 'virny'
DESCRIPTION = "Python library for auditing model stability and fairness"
LONG_DESCRIPTION_CONTENT_TYPE = "text/markdown"
URL = "https://github.com/DataResponsibly/Virny"
EMAILS = "denis.gerasymuk799@gmail.com"
AUTHORS = "Denys Herasymuk"

# The directory containing this file
HERE = os.getcwd()

# Get the long description from the README file
with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

about: dict = {}
with open(os.path.join(HERE, NAME, "__version__.py")) as f:
    exec(f.read(), about)

with pathlib.Path('lib_base_packages.txt').open() as lib_base_packages_txt:
    base_packages = [
        str(requirement)
        for requirement
        in pkg_resources.parse_requirements(lib_base_packages_txt)
    ]

# This call to setup() does all the work
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
    url=URL,
    author=AUTHORS,
    author_email=EMAILS,
    license="BSD-3",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent"
    ],
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    install_requires=base_packages,
    extras_require={
        "dev": base_packages
                + [
                    "pytest~=7.2.1",
                ],
        "docs": [
                    "scikit-learn",
                    "numpy",
                    "scipy",
                    "pandas",
                    "dominate",
                    "flask",
                    "ipykernel",
                    "jupyter-client",
                    "mike",
                    "mkdocs",
                    "mkdocs-awesome-pages-plugin",
                    "mkdocs-material",
                    "mkdocs-redirects",
                    "nbconvert",
                    "python-slugify",
                    "spacy",
                ],
    },
)



"""
Preprocessing techniques.

This module contains function for input dataset preprocessing.
"""
from .basic_preprocessing import preprocess_dataset, get_dummies, make_features_dfs


__all__ = [
    "preprocess_dataset",
    "get_dummies",
    "make_features_dfs",
]



import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

from pprint import pprint
from copy import deepcopy
from datetime import datetime

from sklearn.model_selection import GridSearchCV
from sklearn.metrics import make_scorer, accuracy_score, f1_score
from sklearn.metrics import confusion_matrix, ConfusionMatrixDisplay, classification_report

from virny.custom_classes.base_dataset import BaseFlowDataset


def folds_iterator(n_folds, samples_per_fold, size):
    """
    Iterator for GridSearch based on Cross-Validation

    :param n_folds: number of folds for Cross-Validation
    :param samples_per_fold: number of samples per fold
    """
    for i in range(n_folds):
        yield np.arange(0, size - samples_per_fold * (i + 1)), \
              np.arange(size - samples_per_fold * (i + 1), size - samples_per_fold * i)


def validate_model(model, x, y, params, n_folds, samples_per_fold):
    """
    Use GridSearchCV for a special model to find the best hyperparameters based on validation set
    """
    grid_search = GridSearchCV(estimator=model,
                               param_grid=params,
                               scoring={
                                   "F1_Score": make_scorer(f1_score, average='macro'),
                                   "Accuracy_Score": make_scorer(accuracy_score),
                               },
                               refit="F1_Score",
                               n_jobs=-1,
                               cv=folds_iterator(n_folds, samples_per_fold, x.shape[0]),
                               verbose=10)
    grid_search.fit(x, y.values.ravel())
    best_index = grid_search.best_index_

    return grid_search.best_estimator_, \
           grid_search.cv_results_["mean_test_F1_Score"][best_index], \
           grid_search.cv_results_["mean_test_Accuracy_Score"][best_index], \
           grid_search.best_params_


def test_evaluation(cur_best_model, model_name, cur_best_params,
                    cur_x_train, cur_y_train, cur_x_test, cur_y_test,
                    dataset_title, show_plots, debug_mode):
    """
    Evaluate model on test set.

    :return: F1 score, accuracy and predicted values, which we use to visualisations for model comparison later.
    """
    cur_best_model.fit(cur_x_train, cur_y_train.values.ravel()) # refit model on the whole train set
    cur_model_pred = cur_best_model.predict(cur_x_test)
    test_f1_score = f1_score(cur_y_test, cur_model_pred, average='macro')
    test_accuracy = accuracy_score(cur_y_test, cur_model_pred)

    if debug_mode:
        print("#" * 20, f' {dataset_title} ', "#" * 20)
        print('Test model: ', model_name)
        print('Test model parameters:')
        pprint(cur_best_params)

        # print the scores
        print()
        print(classification_report(cur_y_test, cur_model_pred, digits=3))

    if show_plots:
        # plot the confusion matrix
        sns.set_style("white")
        cm = confusion_matrix(cur_y_test, cur_model_pred, labels=cur_best_model.classes_)
        disp = ConfusionMatrixDisplay(confusion_matrix=cm, display_labels=["Employed", "Not Employed"])
        disp.plot()
        plt.show()

    return test_f1_score, test_accuracy, cur_model_pred


def tune_ML_models(models_params_for_tuning: dict, base_flow_dataset: BaseFlowDataset,
                   dataset_name: str, n_folds: int = 3, samples_per_fold: int = None):
    """
    Tune each model on a validation set with GridSearchCV.

    Return each model with its best hyperparameters that have the highest F1 score and Accuracy.
     results_df is a dataframe with metrics and tuned parameters;
     models_config is a dict with model tuned params for the metrics computation stage
    """
    if samples_per_fold is None:
        samples_per_fold = len(base_flow_dataset.y_test)

    models_config = dict()
    tuned_params_df = pd.DataFrame(columns=('Dataset_Name', 'Model_Name', 'F1_Score', 'Accuracy_Score', 'Model_Best_Params'))
    # Find the most optimal hyperparameters based on accuracy and F1-score for each model in models_config
    for model_idx, (model_name, model_params) in enumerate(models_params_for_tuning.items()):
        try:
            print(f"{datetime.now().strftime('%Y/%m/%d, %H:%M:%S')}: Tuning {model_name}...")
            cur_model, cur_f1_score, cur_accuracy, cur_params = validate_model(deepcopy(model_params['model']),
                                                                               base_flow_dataset.X_train_val,
                                                                               base_flow_dataset.y_train_val,
                                                                               model_params['params'],
                                                                               n_folds, samples_per_fold)
            print(f'{datetime.now().strftime("%Y/%m/%d, %H:%M:%S")}: Tuning for {model_name} is finished '
                  f'[F1 score = {cur_f1_score}, Accuracy = {cur_accuracy}]\n')

        except Exception as err:
            print(f"ERROR with {model_name}: ", err)
            continue

        # Save test results of each model in dataframe
        tuned_params_df.loc[model_idx] = [dataset_name, model_name, cur_f1_score, cur_accuracy, cur_params]
        models_config[model_name] = model_params['model'].set_params(**cur_params)

    return tuned_params_df, models_config


def test_ML_models(best_results_df, models_config, n_folds, X_train, y_train, X_test, y_test,
                   dataset_title, show_plots, debug_mode):
    """
    Find the best model from defined list.
    Tune each model on a validation set with GridSearchCV and
    return best_model with its hyperparameters, which has the highest F1 score
    """
    results_df = pd.DataFrame(columns=('Dataset_Name', 'Model_Name', 'F1_Score',
                                       'Accuracy_Score',
                                       'Model_Best_Params'))
    samples_per_fold = len(y_test)
    best_f1_score = -np.Inf
    best_accuracy = -np.Inf
    best_model_pred = []
    best_model_name = 'No model'
    best_params = None
    idx = 0
    # find the best model among defined in models_config
    for model_config in models_config:
        try:
            print(f"{datetime.now().strftime('%Y/%m/%d, %H:%M:%S')}: Tuning {model_config['model_name']}...")
            cur_model, cur_f1_score, cur_accuracy, cur_params = validate_model(deepcopy(model_config['model']),
                                                                               X_train, y_train, model_config['params'],
                                                                               n_folds, samples_per_fold)
            print(f'{datetime.now().strftime("%Y/%m/%d, %H:%M:%S")}: Tuning for {model_config["model_name"]} is finished')

            test_f1_score, test_accuracy, cur_model_pred = test_evaluation(cur_model, model_config['model_name'], cur_params,
                                                                           X_train, y_train, X_test, y_test, dataset_title, show_plots, debug_mode)
        except Exception as err:
            print(f"ERROR with {model_config['model_name']}: ", err)
            continue

        # save test results of each model in dataframe
        results_df.loc[idx] = [dataset_title,
                               model_config['model_name'],
                               test_f1_score,
                               test_accuracy,
                               cur_params]
        idx += 1

        if test_f1_score > best_f1_score:
            best_f1_score = test_f1_score
            best_accuracy = test_accuracy
            best_model_name = model_config['model_name']
            best_params = cur_params
            best_model_pred = cur_model_pred

    # append results of best model in best_results_df
    best_results_df.loc[best_results_df.shape[0]] = [dataset_title,
                                                     best_model_name,
                                                     best_f1_score,
                                                     best_accuracy,
                                                     best_params,
                                                     best_model_pred]

    return results_df, best_results_df



import pandas as pd


class BaseDataLoader:
    """
    Base data loader class that helps to unify the logic for various datasets.

    Parameters
    ----------
    full_df
        Full dataset in a pandas dataframe format
    target
        Name of the target column name
    numerical_columns
        List of numerical column names
    categorical_columns
        List of categorical column names
    X_data
        [Optional] Dataframe of features
    y_data
        [Optional] Dataframe with a target column
    columns_with_nulls
        [Optional] List of column names that contains nulls

    """
    def __init__(self, full_df: pd.DataFrame, target: str, numerical_columns: list,
                 categorical_columns: list, X_data: pd.DataFrame = None, y_data: pd.DataFrame = None,
                 columns_with_nulls: list = None):
        self.full_df = full_df
        self.target = target
        self.numerical_columns = numerical_columns
        self.categorical_columns = categorical_columns
        self.features = numerical_columns + categorical_columns

        self.X_data = self.full_df[self.features] if X_data is None else X_data
        self.y_data = self.full_df[self.target] if y_data is None else y_data
        self.columns_with_nulls = self.X_data.columns[self.X_data.isna().any().to_list()].to_list() \
            if columns_with_nulls is None else columns_with_nulls
